package provide Brickshelf 1.0

package require http
package require struct
package require htmlparse

namespace eval Brickshelf {
	
# Brickshelf configuration
variable config
array set config {
	
	useragent                "Brickshelf Upload Utility alpha"
	authentication           {}

	login_url                "http://www.brickshelf.com/cgi-bin/gallery.cgi"
	login_username_field     "m"
	login_password_field     "pwd"
	login_cookie             "token"
			
	folder_url               "http://www.brickshelf.com/cgi-bin/gallery.cgi"
	folder_action_field      "action"
	folder_action_value      "newfolder"
	folder_parent_field      "f"
	folder_stdkey_field      "stdkey"
	folder_keywords_field    "keywords"
	folder_name_field        "nf"
	folder_description_field "desc"
	folder_meta_pattern      "URL=http://www\\.brickshelf\\.com/cgi-bin/gallery\\.cgi\\?f=(\\d+)"
	
	upload_url               "http://upload.brickshelf.com/cgi-bin/gallery.cgi"
	upload_action_field      "action"
	upload_action_value      "upload"
	upload_parent_field      "f"
	upload_file_field        "file"
	upload_meta_pattern      "URL=http://www\\.brickshelf\\.com/cgi-bin/gallery\\.cgi\\?f=%s"
	
	crawl_url                "http://www.brickshelf.com/cgi-bin/gallery.cgi?%s=%s"
	crawl_start              "&n=%s"
	crawl_user_mode          "m"
	crawl_folder_mode        "f"
	crawl_link_pattern       "HREF=\"/cgi-bin/gallery\\.cgi\\?f=(\\d+)\""
	crawl_alt_pattern        "ALT=\"(\[^\"\]+)\""
	crawl_next_pattern       "HREF=\"/cgi-bin/gallery\\.cgi\\?(\[mf\])=(\[a-zA-Z0-9\\-_@.\]+)&n=(\\d+)\""

}

# Public procedures
namespace export Authenticate CreateFolder UploadFile

# Package initialization
::http::config -useragent $config(useragent)

#
# Authenticate
#
# http://www.brickshelf.com/cgi-bin/login.cgi
#
# Parameters:
#	username
#	password
#	status, callback that accepts a string argument
#	_inventory, optional name of array to populate with path-number mappings
#
# Return:
#	Brickshelf authentication cookie "key=value"
#
proc Authenticate { username password status {_inventory {}} } {

	variable config
	
	# reset authentication for each login
	set config(authentication) {}
	
	# assemble and issue login query
	eval $status [list "Authenticating..."]
	set login_query [::http::formatQuery \
			$config(login_username_field) $username \
			$config(login_password_field) $password]
	if {[catch {::http::geturl $config(login_url) -query $login_query} token]} {
		return -code error "Authentication failed: $token"
	}
	
	# confirm transfer
	if {[set code [::http::ncode $token]] != 200} {
		::http::cleanup $token
		return -code error "Authentication failed: HTTP $code"
	}
	
	# check authentication cookie
	upvar #0 $token state
	foreach {name value} $state(meta) {
		if {[string equal -nocase "set-cookie" $name]} {
			# the first phrase contains the cookie key=value
			# http://wp.netscape.com/newsref/std/cookie_spec.html
			set cookie [lindex [split $value ";"] 0]
			if {[string equal -nocase $config(login_cookie) [lindex [split $cookie "="] 0]]} {
				set config(authentication) [list "Cookie" $cookie]
				eval $status [list "Authenticated!"]
				break
			}
		}
	}
	
	# login failure is determined by testing for the abscence of an
	# authentication cookie, not by parsing the content of the page.
	if {$config(authentication) == {}} {
		::http::cleanup $token
		return -code error "Authentication failed: invalid login"
	}
		
	# optionally inventory the user's folders
	if {$_inventory != {}} {
				
		# create the inventory array variable
		upvar $_inventory inventory
		if {![array exists inventory]} {
			array set inventory {}
		}
		
		# begin crawling for folders
		eval $status [list "Beginning folder inventory..."]
		if {[catch {FolderCrawl inventory $config(crawl_user_mode) $username "" $status 0 state(body)} err]} {
			return -code error "Authentication inventory failed: $err"
		}
	}
	
	::http::cleanup $token
	return [lindex $config(authentication) 1]
}

# Idea here would be to facilitate an alternate authentication hook
# integrated with the user's existing browser authentication
# (ie, ~/Library/Cookies/Cookies.plist on OS X for Safari)
# proc SetAuthentication { cookie } {
# 	variable config
# 	set config(authentication) [list "Cookie" $cookie]
# }

#
# CreateFolder
#
# Parameters:
#	name
#	description
#	tags, string containing folder keywords (whitespace delimited)
#	status, callback that accepts a string parameter
#	parent, optional numeric ID of parent folder
#		(default 0 for account root)
#
# Return:
#	numeric ID of created folder
#	otherwise error
#
proc CreateFolder { name description tags status {parent 0} } {
	
	variable config
	
	# require login
	if {$config(authentication) == {}} {
		return -code error "Cannot create folder without authentication."
	}
	
	# validate characters in name, description, and tags
	#	names: [a-zA-Z0-9-]; not start or end with -; at least 2 char long
	#	description: presumably same limit
	
	# assemble and issue folder creation query
	eval $status [list "Creating folder $name..."]
	set query [::http::formatQuery \
			$config(folder_action_field) $config(folder_action_value) \
			$config(folder_parent_field) $parent \
			$config(folder_keywords_field) $tags \
			$config(folder_name_field) $name \
			$config(folder_description_field) $description]
	if {[catch {::http::geturl $config(folder_url) -query $query -headers $config(authentication)} token]} {
		return -code error "Folder creation failed: $token"
	}

	# confirm transfer
	if {[set code [::http::ncode $token]] != 200} {
		::http::cleanup $token
		return -code error "Folder creation failed: HTTP $code"
	}
	
	# parse result page
	set doc [::struct::tree]
	if {[catch {::htmlparse::2tree [::http::data $token] $doc} err]} {
		$doc destroy
		::http::cleanup $token
		return -code error "Folder creation failed: cannot parse result"
	}
	
	# brickshelf is wacky; on creation it sends an html http meta redirect as part of the page *body*
	# <META HTTP-EQUIV=Refresh CONTENT="0; URL=http://www.brickshelf.com/cgi-bin/gallery.cgi?f=266689">
	# instead of walking every node, is there an efficient way to simply foreach each "meta" type?
	$doc walk root {node} {
		if {[$doc get $node "type"] eq "meta"} {
			set location [$doc get $node "data"]
			# {URL=http://www\.brickshelf\.com/cgi-bin/gallery\.cgi\?f=(\d+)}
			if {[regexp $config(folder_meta_pattern) $location {} folderid]} {
				if {$folderid ne $parent} {
					set newfolderid $folderid
				}
			}
		}
	}
	
	# was indication of success found?
	if {![info exists newfolderid]} {
		#puts [::http::data $token]
		$doc destroy
		::http::cleanup $token
		return -code error "Folder creation failed: invalid input"
	}
	eval $status [list "Folder created!"]
	
	$doc destroy
	::http::cleanup $token
	return $newfolderid
}

#
# UploadFile
#
# Many thanks to http://wiki.tcl.tk/13789
#
# Parameters:
#	name, local file path (loaded if no _data) 
#	parent, id number of Brickshelf folder to contain uploaded file
#	status, callback that accepts a string parameter
#	_data, optional name of variable containing file data to upload
#		(this mechanism makes it easy to uploaded arbitrary generated
#		content, such as text notes - the remote filename is still
#		determined by this proc's "name" parameter)
#
# Return:
#	none if ok
#	otherwise error
#
proc UploadFile { name parent status {_data {}} } {
	
	variable config
	
	# require login
	if {$config(authentication) == {}} {
		return -code error "Cannot upload without authentication."
	}
	
	set filename [file tail $name]
	
	if {$_data != {}} {
		# refer to the existing data
		upvar $_data filedata
	} else {
		# read the local file
		eval $status [list "Reading $filename..."]
		if {[catch {open $name r} lf]} {
			return -code error "File upload failed: $lf"
		}
		fconfigure $lf -translation binary -encoding binary
		set filedata [read $lf]
		close $lf
	}
	
	# assemble the query
	set qdata ""
	set sep [format "-----NEXTPART_%d_%d" [clock seconds] [pid]]
	append qdata "--$sep\r\nContent-Disposition: form-data; name=\"$config(upload_action_field)\"\r\n\r\n$config(upload_action_value)\r\n"
	append qdata "--$sep\r\nContent-Disposition: form-data; name=\"$config(upload_parent_field)\"\r\n\r\n$parent\r\n"
	append qdata "--$sep\r\nContent-Disposition: form-data; name=\"$config(upload_file_field)\"; filename=\"$filename\"\r\n\r\n$filedata\r\n"
	
	# post the upload
	eval $status [list "Uploading $filename..."]
	if {[catch {::http::geturl $config(upload_url) -type "multipart/form-data; boundary=$sep" \
			-binary true -headers $config(authentication) -query $qdata \
			-queryprogress [list Brickshelf::TransferProgress $status]} token]} {
		return -code error "File upload failed: $token"
	}
	
	#puts [::http::code $token]
	#puts [::http::status $token]
	#puts [::http::error $token]
	
	# confirm transfer
	if {[set code [::http::ncode $token]] != 200} {
		::http::cleanup $token
		return -code error "File upload failed: HTTP $code"
	}
	
	# parse the output
	set doc [::struct::tree]
	if {[catch {::htmlparse::2tree [::http::data $token] $doc} err]} {
		$doc destroy
		::http::cleanup $token
		# has file upload necessarily failed, or just our parsing?
		return -code error "File upload failed: cannot parse results"
	}
	
	# brickshelf is wacky; on upload it sends an html http meta redirect as part of the page body:
	# <META HTTP-EQUIV=Refresh CONTENT="0; URL=http://www.brickshelf.com/cgi-bin/gallery.cgi?f=266689">
	# instead of walking every node, is there an efficient way to simply foreach each "meta" type?
	set confirm 0
	$doc walk root {node} {
		if {[$doc get $node "type"] eq "meta"} {
			set location [$doc get $node "data"]
			set expected [format $config(upload_meta_pattern) $parent]
			if {[regexp $expected $location {}]} {
				set confirm 1
			}
		}
	}
	
	# was indication of success found?
	if {!$confirm} {
		$doc destroy
		::http::cleanup $token
		return -code error "File upload failed: invalid input"
	}
	eval $status [list "Upload complete!"]
	
	$doc destroy
	::http::cleanup $token
}

proc TransferProgress { status token total current } {
	upvar #0 $token state
	if {$current < $total} {
	    eval $status [list [format "Uploading: %.0f%%" [expr {100 * double($current) / double($total)}]]]
	} else {
		eval $status [list "Brickshelf is generating thumbnails..."]
	}
}

#
# FolderCrawl
#
# Recursively populates inventory array with folder numbers
# corresponding to all subfolder paths relative to initial gallery
#
# Parameters
#	_inventory, name of array to populate with path-id mappings for this folder
#	mode, is this a user gallery (m) or a regular folder gallery (f)
#	curfolder, gallery identifier (username for m, numeric id for f)
#	curpath, hierarchical path representation of this folder
#	status, optional callback procedure name of type "proc status { message } { ... }"
#	start, optional parameter to start folder listing after n items
#	_data, optional name of variable containing folder HTML to parse instead of retrieving
#
# Return:
#	none if ok
#	otherwise error
#
proc FolderCrawl { _inventory mode curfolder curpath {status {}} {start 0} {_data {}} } {

	variable config
	upvar $_inventory inventory
	
	set doc [::struct::tree]
	
	# use existing data if available
	if {$_data != {}} {
		
		upvar $_data data
		
		# parse the page data
		if {[catch {::htmlparse::2tree $data $doc} err]} {
			$doc destroy
			return -code error $err
		}
		
	} else {
			
		# what page are we about to retrieve?
		set pageurl [format $config(crawl_url) $mode $curfolder]
		set message $curpath
		if {$start > 0} {
			append pageurl [format $config(crawl_start) $start]
			append message [format " + %s" $start]
		}
		
		# retrieve the page using authentication to access unmoderated folders
		eval $status [list $message]
		if {[catch {::http::geturl $pageurl -headers $config(authentication)} token]} {
			$doc destroy
			return -code error $token
		}
				
		# parse the page
		if {[catch {::htmlparse::2tree [::http::data $token] $doc} err]} {
			$doc destroy
			::http::cleanup $token
			return -code error $err
		}
				
		# release the page
		::http::cleanup $token
	}
		
	# examine each link
 	$doc walk root {node} {
 		if {[$doc get $node "type"] eq "a"} {
 		
 			set location [$doc get $node "data"]
 			
 			# link data: HREF="/cgi-bin/gallery.cgi?i=2634327"
			# original: {^/cgi-bin/gallery\.cgi\?f=(\d+)$}
			# set location [GetAttr [$doc get $node "data"] "href"]
 			
 			if {[regexp $config(crawl_link_pattern) $location {} subfolder]} {
 				
 				# link to subfolder
 				foreach child [$doc children $node] {
 				
 					# Meh. We rely on the alt tags of Brickshelf's folder image links
 					# to determine the actual name of each directory. Not very semantic.
 					if {[$doc get $child "type"] eq "img"} {
 	
 						# img data: BORDER=0 WIDTH=128 HEIGHT=106 ALT="bricksmith-16j-space-search-fix.png" SRC=/gallery/anoved/Examples/thumb/bricksmith-16j-space-search-fix.png_thumb.jpg
 						if {[regexp $config(crawl_alt_pattern) [$doc get $child "data"] {} alt]} {
							
							#set alt [GetAttr [$doc get $child "data"] "alt"]
							set subpath [format "%s/%s" $curpath $alt]
							set inventory($subpath) $subfolder
													
							# look for more folders in this subfolder
							if {[catch {FolderCrawl inventory $config(crawl_folder_mode) $subfolder $subpath $status} err]} {
								$doc destroy
								return -code error $err
							}
						}
 					}
 				}
 				
 			} elseif {[regexp $config(crawl_next_pattern) $location {} nextmode nextfolder next]} {
 				# {^/cgi-bin/gallery\.cgi\?([mf])=([a-zA-Z0-9\-_@\.]+)&n=(\d+)$}
 				
 				# link to next page of this folder
 				if {($nextfolder eq $curfolder) && ($nextmode eq $mode)} {
 					
 					# only follow "next" links, not "prev" links
	 				if {$next > $start} {
	 					
	 					# continue processing contents of next page in same local directory
		 				if {[catch {FolderCrawl inventory $nextmode $nextfolder $curpath $status $next} err]} {
		 					$doc destroy
		 					return -code error $err
		 				}
	 				}
	 			}
 			}
 		}
 	}
	
	$doc destroy
}

}
