package provide BSInventory 1.0

package require Preferences 1.0

namespace eval BSInventory {

# folder path-id inventory
variable folders
array set folders {}

# public procedures
namespace export Reference Reset Count Paths Add ID Save Load Forget

#
# Reference
#
# Parameters:
#	none
#
# Return:
#	name of inventory array
#
proc Reference {} {
	return ::BSInventory::folders
}

#
# Reset
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Reset {} {
	variable folders
	array unset folders
	array set folders {}
}

#
# Count
#
# Parameters:
#	none
#
# Return:
#	number of folders
#
proc Count {} {
	variable folders
	return [array size folders]
}

#
# Paths
#
# Parameters:
#	none
#
# Return:
#	sorted list of folder paths
#
proc Paths {} {
	variable folders
	return [lsort [array names folders]]
}

#
# Add
#
# Parameters:
#	path, new folder path to inventory
#	id, numeric id of new folder path
#
# Return:
#	none
#
proc Add { path id } {
	variable folders
	set folders($path) $id
}

#
# ID
#
# Parameters:
#	folder, optional path to folder
#		(if not specified, account root folder)
#
# Return:
#	numeric ID corresponding to folder (0 for account root)
#	otherwise return error
#
proc ID { {folder "/"} } {
	variable folders
	if {$folder == "/" || $folder == ""} {
		return 0
	} elseif {[info exists folders($folder)]} {
		return $folders($folder)
	} else {
		return -code error "Unrecognized folder: $folderPath"
	}
}

#
# Save
#
# Parameters:
#	none
#
# Return:
#	none
#	error
#
proc Save {} {
	variable folders
	set foldernames [lsort [array names folders]]
	::Preferences::PSet net.anoved.brickshelver folders $foldernames
	foreach foldername $foldernames {
		::Preferences::PSet net.anoved.brickshelver $foldername $folders($foldername)
	}
}

#
# Load
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Load {} {
	variable folders
	::BSInventory::Reset
	foreach foldername [::Preferences::PGet net.anoved.brickshelver folders] {
		set folders($foldername) [::Preferences::PGet net.anoved.brickshelver $foldername]
	}
}

#
# Forget
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Forget {} {
	foreach foldername [::Preferences::PGet net.anoved.brickshelver folders] {
		::Preferences::PDelete net.anoved.brickshelver $foldername
	}
	::Preferences::PDelete net.anoved.brickshelver folders
}

}
