package provide BSLoginWindow 1.0

package require Tk

# wait, what am doing? testing . as the same as
# file dirname info script? Normally it would be
# the same, but it's kind of dumb to try two.
if {[lsearch $auto_path .] == -1} {
	lappend auto_path [file dirname [info script]]
}
package require Brickshelf 1.0
package require BSInventory 1.0
package require Preferences 1.0

namespace eval BSLoginWindow {
	
# external procedure
variable procs
array set procs {
	login {}
}

# login options
variable config
array set config {
	username ""
	password ""
	inventory 1
	remember 0
}

# login state
variable busy 0

# Public procedures
namespace export Setup MakeLoginWindow IsBusy Username

#
# Setup
#
# Parameters:
#	loginProc, procedure to call when authentication is complete
#
# Return:
#	none
#
proc Setup { loginProc } {
	variable procs
	array set procs [list login $loginProc]
}

#
# MakeLoginWindow
#
# Makes the main window a Brickshelf login window.
#
# Parameters
#	none
#
# Return:
#	none
#	error
#
proc MakeLoginWindow {} {
	variable config

	# reset/reload saved folder inventory
	::BSInventory::Load
	
	# apply remembered login settings
	::BSLoginWindow::Load
	
	# force inventory update if no local inventory is present
	set inventorystate normal
	if {[::BSInventory::Count] <= 0} {
		set config(inventory) 1
		set inventorystate disabled
	}
		
	# clear window
	foreach child [winfo children .] {
		destroy $child
	}
	
	# window properties
	wm title . "Brickshelf Login"
	wm geometry . 250x116
	wm minsize . 250 116
	wm resizable . 1 0
			
	# login option: username
	label .usernamel -text "Username:"
	entry .username -bg white -relief sunken -textvariable ::BSLoginWindow::config(username)
	
	# login option: password
	label .passwordl -text "Password:"
	entry .password -bg white -relief sunken -show "*" -textvariable ::BSLoginWindow::config(password)
	
	# login option: update inventory
	checkbutton .inventory -text "Update folder inventory" -variable ::BSLoginWindow::config(inventory) -state $inventorystate
	
	# login option: remember login
	checkbutton .remember -text "Remember login" -variable ::BSLoginWindow::config(remember)
	
	# status bar and login button
	frame .strip
	label .strip.statusl -text "" -anchor w
	button .strip.login -text "Login" -command ::BSLoginWindow::Login
	pack .strip.statusl -side left -fill x -expand true
	pack .strip.login -side right
	
	# keyboard bindings
	bind .username <Return> [list .strip.login invoke]
	bind .password <Return> [list .strip.login invoke]
	bind .strip.login <Return> [list .strip.login invoke]
	
	# grid positioning
	grid .usernamel .username
	grid .passwordl .password
	grid x .inventory
	grid x .remember
	grid .strip -
	
	# grid properties
	grid .username -sticky we
	grid .password -sticky we
	grid .inventory -sticky w
	grid .remember -sticky w
	grid .strip -sticky we
	grid columnconfigure . 1 -weight 1
	
	# select the username
	.username selection range 0 end
	focus .username
}

#
# Login
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Login {} {
	variable config
	variable busy
	variable procs
	
	# indicate login in progress
	set busy 1	
	::BSLoginWindow::DisableLogin
	
	# if and where to perform inventory
	set inventoryname {}
	if {$config(inventory)} {
		::BSInventory::Reset
		set inventoryname [::BSInventory::Reference]
	}
	
	# authenticate (and maybe inventory folders)
	if {[catch {::Brickshelf::Authenticate $config(username) $config(password) ::BSLoginWindow::Status $inventoryname} result]} {
		
		# report login error
		tk_messageBox -type ok -message $result
		
		# remain in login window
		::BSLoginWindow::Status
		::BSLoginWindow::EnableLogin
		set busy 0
		
		return
	}

	# if authentication succeeded, we can save the login information if desired
	if {$config(remember)} {
		::BSLoginWindow::Save
	} else {
		::BSLoginWindow::Forget
	}
	
	# if we updated our inventory, save it (if $config(remember)?)
	if {$config(inventory)} {
		::BSInventory::Save
	}
	
	# proceed to main window
	set busy 0
	$procs(login)
}

#
# DisableLogin
#
# Parameters:
#	none
#
# Return:
#	none
#
proc DisableLogin {} {
	foreach widget {.usernamel .username .passwordl .password .inventory .remember .strip.login} {
		$widget config -state disabled
	}
}

#
# EnableLogin
#
# Parameters:
#	none
#
# Return:
#	none
#
proc EnableLogin {} {
	# this control might not have been enabled to begin with
	if {[::BSInventory::Count] > 0} {
		.inventory config -state normal
	}
	foreach widget {.usernamel .username .passwordl .password .remember .strip.login} {
		$widget config -state normal
	}
}

#
# IsBusy
#
# Parameters:
#	none
#
# Return:
#	1 authentication in progress
#	0 inactive
#
proc IsBusy {} {
	variable busy
	return $busy
}

#
# Username
#
# Parameters:
#	none
#
# Return:
#	username
#
proc Username {} {
	variable config
	return $config(username)
}

#
# Status
#
# Parameters:
#	txt, optional message to display
#
# Return:
#	none
#
proc Status { {txt {}} } {
	.strip.statusl config -text $txt
	update idletasks
}

#
# Save
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Save {} {
	variable config	
	foreach {key value} [array get config] {
		::Preferences::PSet net.anoved.brickshelver $key $value
	}
}

#
# Load
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Load {} {
	variable config
	set u [::Preferences::PGet net.anoved.brickshelver "username"]
	set p [::Preferences::PGet net.anoved.brickshelver "password"]
	set i [::Preferences::PGet net.anoved.brickshelver "inventory"]
	set r [::Preferences::PGet net.anoved.brickshelver "remember"]
 	if {$r eq "1"} {
		array set config [list username $u password $p inventory $i remember 1]
	}
}

#
# Forget
#
# Parameters:
#	none
#
# Return:
#	none
#
proc Forget {} {
	variable config	
	foreach key [array names config] {
		::Preferences::PDelete net.anoved.brickshelver $key
	}
}

}
